function [out,out2,out3,out4,out5,out6,out7] = sgdPhysics_v2(plotRate, stepSizeChoice, estimatorChoice, nIter, userBatch, nOfExperiments, nUser, xi0, nu0,FCStr,UserStr,detThresholdParam,dispMessage)

%% Inputs
% stepSizeChoice: 1 (Decreasing sz), 0 (Fixed sz)
% estimatorChoice: 1 (m_hat), 0 (mu)
% nIter: number of iterations
% userBatch: number of sampled data per user per iteration
% nOfExperiments
% nUser

%% Initialization
% Strategies
% FCStr = [0. 0.4 0.6 0.2];
% UserStr = [0.85 0.95 .25 .5];
% State Transition Matrix
Omega_USER = [UserStr(1) 1-UserStr(1) 0 0; UserStr(2) 1-UserStr(2) 0 0; 0 0 UserStr(3) 1-UserStr(3); 0 0 UserStr(4) 1-UserStr(4)];
Omega_FC = [FCStr(1) 0 1-FCStr(1) 0; 0 FCStr(2) 0 1-FCStr(2); FCStr(3) 0 1-FCStr(3) 0; 0 FCStr(4) 0 1-FCStr(4)];
Omega = Omega_USER*Omega_FC;
l0 = 100; tau0 = .05; % Step Size
emptyList = zeros(nIter/plotRate,1);
averageTrueUserCooperationRate = zeros(nIter/plotRate,1);
expectedUserCooperationRate = zeros(nIter/plotRate,1);
PhiVector = zeros(nIter/plotRate,1);
mNormVector = zeros(nIter/plotRate,1);
rewardVector = zeros(nIter/plotRate,1);
userPayoff = zeros(nIter/plotRate,1);
truePositiveRate = zeros(nIter/plotRate,1);
falsePositiveRate = zeros(nIter/plotRate,1);
detection_tempScore = zeros(nIter/plotRate,1);
frobNormCov = zeros(nIter/plotRate,1);
maxEigenCov = zeros(nIter/plotRate,1);
determinantEigenCov = zeros(nIter/plotRate,1);
sizeCovMatrix = zeros(nIter/plotRate,1);
for experiment = 1:nOfExperiments
    % Noise Levels (nu: Privacy protection, xi: Defective Action)
    % xi0 = 1.5; nu0 = 0.5;
    
    % Initial State Distribution
    lambda = [1 0 0 0];
    actionUser = ones(nUser,1);
    actionFC = ones(nUser,1);
    
    disp(['Exp:' mat2str(experiment) ' est ' mat2str(estimatorChoice) ' nUser ' mat2str(nUser) ' Str ' dispMessage] )
    %% SGD Initializations
    load('physics_train.mat');
    A = A(randperm(size(A,1)),:);
    % Separate the features and the classes
    Zf = A(:,1:end-1);
    Zc = A(:,end);
    % Set the classes indexes to be {−1,+1}.
    Zc = 2*Zc-1;
    % Remove empty features
    I = find(mean(abs(Zf))>0.1); 
    Zf = Zf(:,I);
    [nOfSamples,nOfDimensions] = size(Zf);
    % Normalize Zf
    Zf = Zf-repmat(mean(Zf),[size(Zf,1),1]);
    Zf = Zf ./ repmat( sqrt(sum(Zf.^2)/size(Zf,1)), [size(Zf,1),1] );
    % Gradient Function Description
    nablaLL = @(ell)1./(1+exp(-ell));
    nablaFTilde = @(theta,i)-Zc(i)*Zf(i,:)'*nablaLL(-Zc(i)*(Zf(i,:)*theta));
    
    % Objective Function
    L = @(s,Z_y)1/nOfSamples * sum( log( 1 + exp(-s.*Z_y) ) );
    E = @(param,Z_x,Z_y)L(Z_x*param,Z_y);
    
    % Optimization Parameter
    theta = 2*rand(nOfDimensions,1);
    
    ElistS = zeros(nIter/plotRate,1);
    
    %% Outputs for Analysis
    trueUserCooperationRate = zeros(nIter/plotRate,1);
    phiVectorIteration = zeros(nIter/plotRate,1);
    mNormIteration = zeros(nIter/plotRate,1);
    tempRewardVector = zeros(nIter/plotRate,1);
    tempUserPayoffVector = zeros(nIter/plotRate,1);
    tempTruePositiveRateVector = zeros(nIter/plotRate,1);
    tempFalsePositiveRateVector = zeros(nIter/plotRate,1);
    tempDetection_tempScore = zeros(nIter/plotRate,1);
    temp_frobNormCov = zeros(nIter/plotRate,1);
    temp_maxEigenCov = zeros(nIter/plotRate,1);
    temp_determinantEigenCov= zeros(nIter/plotRate,1);
    temp_sizeCovMatrix = zeros(nIter/plotRate,1);
    
    %% SGD Iterations
    for t = 1:nIter
        
        % Reduce xi and nu
%         xiStep = 100;
%         nuStep = 100;
%         xi = xi0/(1+t/xiStep);
%         nu = nu0/(1+t/nuStep);
        xi = xi0;
        nu = nu0;
        
        X = zeros(nOfDimensions, nUser);
        Y = zeros(nOfDimensions, nUser);
        
        %% User Side
        tempTrueUserCooperationRate = 0;
        dataIndex = 1 + floor(rand(userBatch,nUser)*nOfSamples);
        for k = 1:nUser
            % Stochastic Gradients are Generated
            x_k = zeros(nOfDimensions,1);
            for userDataIndex = 1:userBatch
                x_k = x_k + nablaFTilde(theta,dataIndex(userDataIndex,k));
            end
            X(:,k) = x_k/userBatch;
            % User Actions are generated
            if actionFC(k,1) == 1 %C
                if actionUser(k,1) == 1 %c
                    StrategyUserTemp = UserStr(1);
                else
                    StrategyUserTemp = UserStr(2);
                end
            else
                if actionUser(k,1) == 1
                    StrategyUserTemp = UserStr(3);
                else
                    StrategyUserTemp = UserStr(4);
                end
            end
            actionUser(k,1) = rand < StrategyUserTemp;
            % User Reports are generated
            if actionUser(k,1) == 1
                Y(:,k) = X(:,k) + mvnrnd(zeros(nOfDimensions,1),(nu^2)*eye(nOfDimensions),1)';
                tempTrueUserCooperationRate = tempTrueUserCooperationRate + 1;
            else
                Y(:,k) = mvnrnd(zeros(nOfDimensions,1),(xi^2)*eye(nOfDimensions),1)';
            end
        end
                
        %% FC Joint Gradient Est and Hypothesis Testing
        denom = nUser*lambda*UserStr';
        m_hat = sum(Y,2)/denom;
        m_Norm = transpose(m_hat)*m_hat;
        mu = zeros(nOfDimensions,1);
        mu_d = zeros(nOfDimensions,1);
        coopUserPrediction = 0;
        truePositive = 0;
        falsePositive = 0;
        detThreshold = detThresholdParam*log((1-lambda*UserStr')/(lambda*UserStr'));
        for k = 1:nUser
            belief = (transpose(Y(:,k)))*m_hat > 0.5*m_Norm + detThreshold;
            if belief == 1 %c_hat
                mu = mu + Y(:,k);
                coopUserPrediction  = coopUserPrediction + 1;
                if actionUser(k,1) == 1
                    truePositive = truePositive + 1;
                else
                    falsePositive = falsePositive + 1;
                end 
            else
                mu_d = mu_d + Y(:,k);
            end
        end
        mu = mu/coopUserPrediction;
        mu_d = mu_d/(nUser-coopUserPrediction);
        covMatrix = zeros(nOfDimensions,nOfDimensions);
        covMatrix_d = zeros(nOfDimensions,nOfDimensions);
        for k = 1:nUser
            belief = (transpose(Y(:,k)))*m_hat > 0.5*m_Norm + detThreshold;
            if belief == 1 %c_hat
                covMatrix = covMatrix + (Y(:,k)-mu)*(transpose(Y(:,k))-transpose(mu));
            else
                covMatrix_d = covMatrix_d + (Y(:,k)-mu_d)*(transpose(Y(:,k))-transpose(mu_d));
            end        
        end
        
        %% Temporary Modification
        % Remove empty features
        I_Y = find(mean(abs(transpose(Y)))>0); 
        Y_mod = Y(I_Y,:);   
        covMatrix_temp = cov(transpose(Y_mod));
        mean_temp = mean(transpose(Y_mod));
        tempScore = mean_temp*inv(covMatrix_temp)*transpose(mean_temp);
        temp_frobNorm = norm(covMatrix_temp,'fro');
        temp_maxEigen = norm(covMatrix_temp);
        temp_determinant = det(covMatrix_temp);
        % temp_covMatrixSize = length(I_Y);
            temp_covMatrixSize = sum(svd(covMatrix_temp)>0.0001);
        %% End of Temporary Modification
        
        covMatrix = covMatrix/(coopUserPrediction-1);
        denomPhi = sqrt(transpose(mu)*covMatrix*mu);
        numPhi = transpose(m_hat)*(mu-0.5*m_hat-detThreshold);
        Phi = 1- qfunc(numPhi/(denomPhi));
        
        covMatrix_d = covMatrix_d/(nUser - coopUserPrediction-1);
        denomPsi = sqrt(transpose(mu_d)*covMatrix_d*mu_d);
        numPsi = transpose(m_hat)*(-mu_d+0.5*m_hat+detThreshold);
        Psi = qfunc(numPsi/(denomPsi));
        
        pi1 = (FCStr(1)*(1-Psi) - FCStr(2)*(1-Phi))/(Phi-Psi);
        pi2 = (FCStr(2)*Phi - FCStr(1)*Psi)/(Phi-Psi);
        pi3 = (FCStr(3)*(1-Psi) - FCStr(4)*(1-Phi))/(Phi-Psi);
        pi4 = (FCStr(4)*Phi - FCStr(3)*Psi)/(Phi-Psi);
        

        
        %% FC Actions are Generated
        tempReward = 0;
        for k = 1:nUser
            belief = (transpose(Y(:,k)))*m_hat > 0.5*m_Norm + detThreshold;
            if actionFC(k,1) == 1 %C
                if belief == 1 %c_hat
                    StrategyFCTemp = pi1;
                else
                    StrategyFCTemp = pi2;
                end
            else
                if belief == 1
                    StrategyFCTemp = pi3;
                else
                    StrategyFCTemp = pi4;
                end
            end
            ActionFCTemp = rand < StrategyFCTemp;
            actionFC(k,1) = ActionFCTemp;
            tempReward = tempReward + ActionFCTemp;
        end
        
        if mod(t,plotRate)==1
            ElistS( 1+(t-1)/plotRate,1) = E(theta,Zf,Zc);
            trueUserCooperationRate(1+(t-1)/plotRate,1) = tempTrueUserCooperationRate/nUser;
            expectedUserCooperationRate(1+(t-1)/plotRate,1) = lambda*UserStr';
            phiVectorIteration(1+(t-1)/plotRate,1) = Phi;
            mNormIteration(1+(t-1)/plotRate,1) = sqrt(m_Norm);
            tempRewardVector(1+(t-1)/plotRate,1) = sum(actionFC)/nUser;
            tempUserPayoffVector(1+(t-1)/plotRate,1) = sum(actionFC-0.5*actionUser)/nUser;
            tempTruePositiveRateVector(1+(t-1)/plotRate,1) = truePositive/sum(actionUser);
            tempFalsePositiveRateVector(1+(t-1)/plotRate,1) = falsePositive/(nUser-sum(actionUser));
           tempDetection_tempScore(1+(t-1)/plotRate,1) = tempScore;
            temp_frobNormCov(1+(t-1)/plotRate,1) = temp_frobNorm;
            temp_maxEigenCov(1+(t-1)/plotRate,1) = temp_maxEigen;
            temp_determinantEigenCov(1+(t-1)/plotRate,1) = temp_determinant;
            temp_sizeCovMatrix(1+(t-1)/plotRate,1) = temp_covMatrixSize;
        end
        
        if stepSizeChoice == 1
            tau = tau0/(1+t/l0);
        else
            tau = 0.01;
        end
        if estimatorChoice == 1
            theta = theta - tau*m_hat;
        else
            theta = theta - tau*mu;
        end
        lambda=(lambda*Omega);
    end
%u = log10(ElistS-0.543121286203416);
%plot(1:plotRate:nIter, u, 'b');
% plot(1:plotRate:nIter,trueUserCooperationRate, 'b')
emptyList = emptyList + ElistS;
averageTrueUserCooperationRate = averageTrueUserCooperationRate + trueUserCooperationRate;
PhiVector = PhiVector + phiVectorIteration;
mNormVector = mNormVector + mNormIteration;
userPayoff = userPayoff + tempUserPayoffVector;
rewardVector = rewardVector + tempRewardVector;
truePositiveRate = tempTruePositiveRateVector + truePositiveRate;
falsePositiveRate = tempFalsePositiveRateVector + falsePositiveRate;
detection_tempScore = detection_tempScore + tempDetection_tempScore;
frobNormCov = frobNormCov + temp_frobNormCov;
maxEigenCov = maxEigenCov + temp_maxEigenCov;
determinantEigenCov = determinantEigenCov + temp_determinantEigenCov;
sizeCovMatrix = sizeCovMatrix + temp_sizeCovMatrix;
end

ElistS = emptyList/nOfExperiments;
u = log10(ElistS-0.543121286203416);
figure, plot(1:plotRate:nIter, u, 'b');
title(['log(E(w_l) - min E) q='  mat2str(UserStr)  ' p='  mat2str(FCStr) ]); set(gca, 'FontSize', 10);

averageTrueUserCooperationRate = averageTrueUserCooperationRate/nOfExperiments;
figure, plot(1:plotRate:nIter, averageTrueUserCooperationRate, 'b', 1:plotRate:nIter, expectedUserCooperationRate, 'r');
title(['User Cooperation Probability nu=' mat2str(nu0) ' p=' mat2str(FCStr)]); set(gca, 'FontSize', 10);

PhiVector = PhiVector/nOfExperiments;
figure, plot(1:plotRate:nIter, PhiVector, 'b');
title(['Phi q='  mat2str(UserStr)  ' p=' mat2str(FCStr)]); set(gca, 'FontSize', 10);

mNormVector = mNormVector/nOfExperiments;
figure, plot(1:plotRate:nIter, mNormVector, 'b');
title(['mNorm q='  mat2str(UserStr)  ' p=' mat2str(FCStr)]); set(gca, 'FontSize', 10);

userPayoff = userPayoff/nOfExperiments;
figure, plot(1:plotRate:nIter, userPayoff, 'b');
title(['userPayoff q=' mat2str(UserStr) ' p=' mat2str(FCStr)]); set(gca, 'FontSize', 10);

rewardVector = rewardVector/nOfExperiments;
figure, plot(1:plotRate:nIter, rewardVector, 'b');
title(['Reward Vector q='  mat2str(UserStr)  ' p=' mat2str(FCStr)]); set(gca, 'FontSize', 10);

truePositiveRate = truePositiveRate/nOfExperiments;
figure, plot(1:plotRate:nIter, truePositiveRate, 'b');
title(['True Positive Rate q='  mat2str(UserStr)  ' p=' mat2str(FCStr)]); set(gca, 'FontSize', 10);

falsePositiveRate = falsePositiveRate/nOfExperiments;
figure, plot(1:plotRate:nIter, falsePositiveRate, 'b');
title(['False Positive Rate q='  mat2str(UserStr)  ' p=' mat2str(FCStr)]); set(gca, 'FontSize', 10);

% detection_tempScore = detection_tempScore/nOfExperiments;
% figure, plot(1:plotRate:nIter, detection_tempScore, 'b');
% title(['mTInv(Cov)m q='  mat2str(UserStr)  ' p=' mat2str(FCStr)]); set(gca, 'FontSize', 10);
% 
% frobNormCov = frobNormCov/nOfExperiments;
% maxEigenCov = maxEigenCov/nOfExperiments;
% figure, plot(1:plotRate:nIter, frobNormCov, 'b', 1:plotRate:nIter, maxEigenCov, 'r');
% title(['Cov Matrix Norms q='  mat2str(UserStr)  ' p=' mat2str(FCStr)]); set(gca, 'FontSize', 10);
% legend('Frobenius Norm', 'Spectral Norm');
% 
% determinantEigenCov = determinantEigenCov/nOfExperiments;
% figure, plot(1:plotRate:nIter, determinantEigenCov, 'b');
% title(['Cov Matrix Determinant q='  mat2str(UserStr)  ' p=' mat2str(FCStr)]); set(gca, 'FontSize', 10);
% 
% sizeCovMatrix = sizeCovMatrix/nOfExperiments;
% figure, plot(1:plotRate:nIter, sizeCovMatrix, 'b');
% title(['Dim Cov Matrix q='  mat2str(UserStr)  ' p=' mat2str(FCStr)]); set(gca, 'FontSize', 10);

out = u;
out2 = truePositiveRate;
out3 = falsePositiveRate;
out4 = averageTrueUserCooperationRate;
out5 = mNormVector;
out6 = rewardVector;
out7 = userPayoff;