import argparse
import os


def parse_args():
    parser = argparse.ArgumentParser(description="Pytorch implementation.")
    folder_text = 'simulations_gap_1'
    parser.add_argument('--base_folder', type=str, default=folder_text + '/saved_baseline')
    parser.add_argument('--save_folder', type=str, default=folder_text + '/data_folder')
    parser.add_argument('--agent_num', type=int, default=50, help='Number of users in the network')
    parser.add_argument('--batch_size', type=int, default=8)
    parser.add_argument('--test_batch_size', type=int, default=128)
    parser.add_argument('--experiment', type=str, default='with_grad_trac')  # with_grad_trac, without_grad_trac
    parser.add_argument('--user_model', type=str, default='same')  # same, random
    parser.add_argument('--model_init', type=str, default='zero')  # zero, random
    parser.add_argument('--metric', type=str, default='gap')  # gap, vanilla
    parser.add_argument('--data_dim', type=int, default=784)  # 784 for MNIST
    parser.add_argument('--rho', type=float, default=0.1)
    parser.add_argument('--cuda', type=bool, default=False)
    parser.add_argument('--save_model', type=bool, default=False)
    parser.add_argument('--cuda_index', type=int, default=0)
    parser.add_argument('--learning_rate', type=float, default=0.1)
    parser.add_argument('--link_prob', type=float, default=1)
    parser.add_argument('--iter_num', type=int, default=201)
    parser.add_argument('--workers', type=int, default=1)
    parser.add_argument('--save_interval', type=int, default=1)
    parser.add_argument('--data_folder', type=str, default='MNIST_all_data')
    parser.add_argument('--class_num', type=int, default=10)  # number of classes in the selected dataset
    parser.add_argument('--num_of_samples', type=int, default=10000)  # dataset size at each agent
    parser.add_argument('--test_num_of_samples', type=int, default=5000)  # dataset size for testing
    # test_num_of_samples + num_of_samples = 50000
    parser.add_argument('--regression_type', type=str, default='quadratic')  # linear, logistic
    parser.add_argument('--fcStr', type=list, default=[0.95, 0.75, 0.2, 0])
    parser.add_argument('--userStr', type=list, default=[0.9, 0.9, 0.9, 0.9])
    parser.add_argument('--lambda_init', type=list, default=[1, 0, 0, 0])
    parser.add_argument('--xi', type=float, default=1)
    parser.add_argument('--nu', type=float, default=0.001)
    parser.add_argument('--robust_selection', type=str, default='robust')  # robust, gaussian

    if not os.path.exists(folder_text):
        os.mkdir(folder_text)

    args = parser.parse_args()

    return args


def save_module(arg, save_fld, iters, save_vector, vector_content):
    for i in range(len(vector_content)):
        if vector_content[i] == "training_loss":
            file = save_fld + '/training_loss.txt'
            gamma_save = save_vector[i]
            if not os.path.exists(file):
                f = open(file, "w+")
                f.write(str(gamma_save))
                f.close()
            else:
                f = open(file, "a")
                f.write(','+str(gamma_save))
                f.close()
        elif vector_content[i] == "testing_loss":
            file = save_fld + '/testing_loss.txt'
            gamma_save = save_vector[i]
            if not os.path.exists(file):
                f = open(file, "w+")
                f.write(str(gamma_save))
                f.close()
            else:
                f = open(file, "a")
                f.write(','+str(gamma_save))
                f.close()
        elif vector_content[i] == "training_accuracy":
            file = save_fld + '/training_accuracy.txt'
            gamma_save = save_vector[i]
            if not os.path.exists(file):
                f = open(file, "w+")
                f.write(str(gamma_save))
                f.close()
            else:
                f = open(file, "a")
                f.write(','+str(gamma_save))
                f.close()
        elif vector_content[i] == "testing_accuracy":
            file = save_fld + '/testing_accuracy.txt'
            gamma_save = save_vector[i]
            if not os.path.exists(file):
                f = open(file, "w+")
                f.write(str(gamma_save))
                f.close()
            else:
                f = open(file, "a")
                f.write(','+str(gamma_save))
                f.close()
        elif vector_content[i] == "pFA":
            file = save_fld + '/pFA.txt'
            gamma_save = save_vector[i]
            if not os.path.exists(file):
                f = open(file, "w+")
                f.write(str(gamma_save))
                f.close()
            else:
                f = open(file, "a")
                f.write(','+str(gamma_save))
                f.close()
        elif vector_content[i] == "pDet":
            file = save_fld + '/pDet.txt'
            gamma_save = save_vector[i]
            if not os.path.exists(file):
                f = open(file, "w+")
                f.write(str(gamma_save))
                f.close()
            else:
                f = open(file, "a")
                f.write(','+str(gamma_save))
                f.close()

    file = save_fld + '/iterations.txt'
    if not os.path.exists(file):
        f = open(file, "w+")
        f.write(str(iters))
        f.close()
    else:
        f = open(file, "a")
        f.write(',' + str(iters))
        f.close()


def parse_args_fid():
    parser = argparse.ArgumentParser(description="Pytorch implementation.")
    parser.add_argument('--window', type=int, default=1)
    parser.add_argument('--pick_data_per', type=int, default=1)

    args = parser.parse_args()

    file = open("figure_setup_3.txt", "r")  # figure_setup.txt, figure_setup_cum_loss.txt
    labels = []
    legend_text = []
    iteration_files = []
    fid_files = []
    k = 0
    file = file.read().splitlines()
    for line in file:
        fields = line.split(";")
        if k == 0:
            labels.append(fields[0])
            labels.append(fields[1])
            labels.append(fields[2])
        elif k == 1:
            legend_text = fields
        elif k == 2:
            iteration_files = fields
        elif k == 3:
            fid_files = fields
        k += 1
    setattr(args, 'labels', labels)
    setattr(args, 'legend_text', legend_text)
    setattr(args, 'iteration_files', iteration_files)
    setattr(args, 'fid_files', fid_files)

    return args
