import numpy as np
import torch
from torch.utils.data import TensorDataset, DataLoader
import glob
import os
from PIL import Image


class data_provider(object):
    def __init__(self, args, train_='regular'):
        self.args = args
        self.files = glob.glob(args.data_folder + '/*')
        file_idx = np.arange(len(self.files))
        if train_ == 'regular':
            permuted_idx = np.random.permutation(file_idx[:50000 + args.test_num_of_samples])
            self.selected_file_idx = permuted_idx[:args.num_of_samples]
        elif train_ == 'train':
            permuted_idx = np.random.permutation(file_idx[:50000 + args.test_num_of_samples])
            self.selected_file_idx = permuted_idx
        elif train_ == 'test':
            self.selected_file_idx = file_idx[args.test_num_of_samples:]
        new_file = []
        new_label = []
        for k in self.selected_file_idx:
            label = int(os.path.split(self.files[k])[-1].split('_')[-2])
            data = np.asarray(Image.open(self.files[k]), dtype="float32")
            data = data.reshape(1, 28, 28)
            data = (data - 127.5) / 127.5  # Normalize the images to [-1, 1]
            new_file.append(data)
            new_label.append(label)
            # print(str(k))
        x_data = np.array(new_file)
        y_data = np.array(new_label)

        tensor_x = torch.Tensor(x_data)
        tensor_y = torch.LongTensor(y_data)
        my_dataset = TensorDataset(tensor_x, tensor_y)

        if train_ == 'regular':
            self.train_ds = get_infinite_batches(torch.utils.data.DataLoader(my_dataset,
                                                                             batch_size=self.args.batch_size,
                                                                             shuffle=True,
                                                                             num_workers=self.args.workers,
                                                                             pin_memory=True))
        else:
            self.train_ds = get_infinite_batches(torch.utils.data.DataLoader(my_dataset,
                                                                             batch_size=self.args.test_batch_size,
                                                                             shuffle=True,
                                                                             num_workers=self.args.workers,
                                                                             pin_memory=True))


    def get_data(self):
        next_pair = self.train_ds.__next__()
        u, v = next_pair[0], next_pair[1]
        return u, v


def get_infinite_batches(iterator_obj):
    while True:
        for i, (samples_) in enumerate(iterator_obj):
            yield samples_
