import torch
from torch.autograd import Variable
from DNN import DNN_v1
import torch.nn as nn
import numpy as np
from scipy import special
from scipy.spatial.distance import cdist, euclidean


class learning_model(object):
    def __init__(self, args):
        print("Creating a new learning model...")
        self.cuda_index = args.cuda_index
        self.cuda = args.cuda
        self.DNN_model = DNN_v1()
        self.batch_size = args.batch_size
        self.workers = args.workers
        self.loss_fn = nn.CrossEntropyLoss()
        self.eta = args.learning_rate
        self.user_num = args.agent_num
        self.check_cuda(self.cuda)
        self.xi = args.xi
        self.nu = args.nu
        self.args = args

    def check_cuda(self, cuda_flag=False):
        print(cuda_flag)
        if cuda_flag:
            self.cuda_index = 0
            self.cuda = True
            self.DNN_model.cuda(self.cuda_index)
            print("Models are assigned to GPU : {}".format(self.cuda_index))
        else:
            self.cuda = False

    def grad_compute(self, samples, labels, iteration, agent, actionUser):
        if actionUser == 1:
            torch_samples = self.get_torch_variable(samples)
            torch_labels = self.get_torch_variable(labels)
            torch_output = self.DNN_model(torch_samples)
            torch_cost = self.loss_fn(torch_output, torch_labels)

            cost = torch_cost.clone().detach().cpu().numpy()
            # print('Loss at iteration ', str(iteration), ' for agent ', str(agent), ' is: ', str(cost))
            torch_cost.backward()

            params = list(self.DNN_model.parameters())
            gradients_ = []
            linear_grads = []
            with torch.no_grad():
                for param in params:
                    gradients = param.grad.clone()
                    shape_grad = list(gradients.shape)
                    if len(shape_grad) == 1:
                        priv_noise = np.sqrt(self.nu) * np.random.randn(shape_grad[0])
                    elif len(shape_grad) == 2:
                        priv_noise = np.sqrt(self.nu) * np.random.randn(shape_grad[0], shape_grad[1])
                    elif len(shape_grad) == 3:
                        priv_noise = np.sqrt(self.nu) * np.random.randn(shape_grad[0], shape_grad[1],
                                                                        shape_grad[2])
                    else:
                        priv_noise = np.sqrt(self.nu) * np.random.randn(shape_grad[0], shape_grad[1],
                                                                        shape_grad[2], shape_grad[3])
                    gradients += torch.Tensor(priv_noise)
                    gradients_.append(gradients)
                    # Linearization
                    temp_vector = gradients.cpu().numpy()
                    initial_shape = temp_vector.shape
                    new_shape = 1
                    for j in range(len(initial_shape)):
                        new_shape = new_shape * initial_shape[j]
                    final_vector = temp_vector.reshape(new_shape)
                    linear_grads.extend(final_vector)

                    new_grad_val = torch.zeros(gradients.shape)
                    param.grad.copy_(new_grad_val)
        else:
            torch_samples = self.get_torch_variable(samples)
            torch_labels = self.get_torch_variable(labels)
            torch_output = self.DNN_model(torch_samples)
            torch_cost = self.loss_fn(torch_output, torch_labels)

            cost = torch_cost.clone().detach().cpu().numpy()
            # print('Loss at iteration ', str(iteration), ' for agent ', str(agent), ' is: ', str(cost))

            params = list(self.DNN_model.parameters())
            gradients_ = []
            linear_grads = []
            with torch.no_grad():
                for param in params:
                    shape_grad = list(param.shape)
                    if len(shape_grad) == 1:
                        def_noise = np.sqrt(self.xi) * np.random.randn(shape_grad[0])
                    elif len(shape_grad) == 2:
                        def_noise = np.sqrt(self.xi) * np.random.randn(shape_grad[0], shape_grad[1])
                    elif len(shape_grad) == 3:
                        def_noise = np.sqrt(self.xi) * np.random.randn(shape_grad[0], shape_grad[1],
                                                                       shape_grad[2])
                    else:
                        def_noise = np.sqrt(self.xi) * np.random.randn(shape_grad[0], shape_grad[1],
                                                                       shape_grad[2], shape_grad[3])
                    # Linearization
                    temp_vector = def_noise
                    initial_shape = temp_vector.shape
                    new_shape = 1
                    for j in range(len(initial_shape)):
                        new_shape = new_shape * initial_shape[j]
                    final_vector = temp_vector.reshape(new_shape)
                    linear_grads.extend(final_vector)

                    tensor_def_noise = torch.Tensor(def_noise)
                    gradients_.append(tensor_def_noise)
        return gradients_, cost, np.array(linear_grads)

    def grad_compute_robust(self, samples, labels, iteration, agent, actionUser):
        if actionUser == 1:
            torch_samples = self.get_torch_variable(samples)
            torch_labels = self.get_torch_variable(labels)
            torch_output = self.DNN_model(torch_samples)
            torch_cost = self.loss_fn(torch_output, torch_labels)

            cost = torch_cost.clone().detach().cpu().numpy()
            # print('Loss at iteration ', str(iteration), ' for agent ', str(agent), ' is: ', str(cost))
            torch_cost.backward()

            params = list(self.DNN_model.parameters())
            gradients_ = []
            linear_grads = []
            with torch.no_grad():
                for param in params:
                    gradients = param.grad.clone()
                    shape_grad = list(gradients.shape)
                    if len(shape_grad) == 1:
                        priv_noise = np.sqrt(self.nu) * np.random.randn(shape_grad[0])
                    elif len(shape_grad) == 2:
                        priv_noise = np.sqrt(self.nu) * np.random.randn(shape_grad[0], shape_grad[1])
                    elif len(shape_grad) == 3:
                        priv_noise = np.sqrt(self.nu) * np.random.randn(shape_grad[0], shape_grad[1],
                                                                        shape_grad[2])
                    else:
                        priv_noise = np.sqrt(self.nu) * np.random.randn(shape_grad[0], shape_grad[1],
                                                                        shape_grad[2], shape_grad[3])
                    gradients += torch.Tensor(priv_noise)
                    gradients_.append(gradients)
                    # Linearization
                    temp_vector = gradients.cpu().numpy()
                    initial_shape = temp_vector.shape
                    new_shape = 1
                    for j in range(len(initial_shape)):
                        new_shape = new_shape * initial_shape[j]
                    final_vector = temp_vector.reshape(new_shape)
                    linear_grads.extend(final_vector)

                    new_grad_val = torch.zeros(gradients.shape)
                    param.grad.copy_(new_grad_val)
        else:
            torch_samples = self.get_torch_variable(samples)
            torch_labels = self.get_torch_variable(labels)
            torch_output = self.DNN_model(torch_samples)
            torch_cost = self.loss_fn(torch_output, torch_labels)

            cost = torch_cost.clone().detach().cpu().numpy()
            # print('Loss at iteration ', str(iteration), ' for agent ', str(agent), ' is: ', str(cost))
            torch_cost.backward()

            params = list(self.DNN_model.parameters())
            gradients_ = []
            linear_grads = []
            with torch.no_grad():
                for param in params:
                    gradients = param.grad.clone()
                    shape_grad = list(gradients.shape)
                    if len(shape_grad) == 1:
                        priv_noise = np.sqrt(self.nu) * np.random.randn(shape_grad[0])
                    elif len(shape_grad) == 2:
                        priv_noise = np.sqrt(self.nu) * np.random.randn(shape_grad[0], shape_grad[1])
                    elif len(shape_grad) == 3:
                        priv_noise = np.sqrt(self.nu) * np.random.randn(shape_grad[0], shape_grad[1],
                                                                        shape_grad[2])
                    else:
                        priv_noise = np.sqrt(self.nu) * np.random.randn(shape_grad[0], shape_grad[1],
                                                                        shape_grad[2], shape_grad[3])
                    gradients += torch.Tensor(priv_noise)
                    gradients = -gradients
                    gradients_.append(gradients)
                    # Linearization
                    temp_vector = gradients.cpu().numpy()
                    initial_shape = temp_vector.shape
                    new_shape = 1
                    for j in range(len(initial_shape)):
                        new_shape = new_shape * initial_shape[j]
                    final_vector = temp_vector.reshape(new_shape)
                    linear_grads.extend(final_vector)

                    new_grad_val = torch.zeros(gradients.shape)
                    param.grad.copy_(new_grad_val)
        return gradients_, cost, np.array(linear_grads)

    def get_torch_variable(self, inp):
        if self.cuda:
            return Variable(inp).cuda(self.cuda_index)
        else:
            return Variable(inp)

    def update_params(self, grad_, rho, lin_grad=None):
        m_hat = None
        if self.args.robust_selection == 'robust':
            m_hat = geoMed(lin_grad)
            current_ind = 0
            with torch.no_grad():
                params = list(self.DNN_model.parameters())
                for m in range(len(params)):
                    # Reshape into original shape
                    new_shape = 1
                    initial_shape = params[m].shape
                    for j in range(len(initial_shape)):
                        new_shape = new_shape * initial_shape[j]
                    final_m_hat = m_hat[current_ind:current_ind + new_shape].reshape(initial_shape)
                    final_m_hat_tensor = torch.from_numpy(np.float32(final_m_hat))
                    current_ind += new_shape
                    new_model_val = params[m].clone() - self.eta * final_m_hat_tensor
                    params[m].copy_(new_model_val)
        else:
            with torch.no_grad():
                params = list(self.DNN_model.parameters())
                for m in range(len(params)):
                    sum_ = 0
                    for i in range(len(grad_)):
                        sum_ += grad_[i][m]
                    new_model_val = params[m].clone() - self.eta / self.user_num / rho * sum_
                    params[m].copy_(new_model_val)
        return m_hat

    def check_accuracy(self, batch_test, labels_test):
        self.DNN_model.eval()
        torch_samples = self.get_torch_variable(batch_test)
        torch_labels = self.get_torch_variable(labels_test)
        torch_output = self.DNN_model(torch_samples)
        # torch_cost = sum(torch_output) / self.batch_size
        torch_cost = self.loss_fn(torch_output, torch_labels)
        cost = torch_cost.clone().detach().cpu().numpy()
        # print('Loss at iteration ', str(iteration), ' is: ', str(cost))

        correct = 0
        total = 0
        _, predicted = torch.max(torch_output.data, 1)
        total += torch_labels.size(0)
        correct += (predicted == torch_labels).sum().item()
        accuracy = correct / total
        # print('Accuracy at iteration ', str(iteration), ' is: ', str(accuracy))

        return cost, accuracy


def update_action_user(actionUser, actionFc, userStr):
    for i in range(len(actionUser)):
        if actionUser[i] == 1:
            if actionFc[i] == 1:
                temp = userStr[0]
            else:
                temp = userStr[2]
        else:
            if actionFc[i] == 1:
                temp = userStr[1]
            else:
                temp = userStr[3]
        rand_num = np.random.rand(1)
        if rand_num < temp:
            actionUser[i] = 1
        else:
            actionUser[i] = 0
    return actionUser


def update_action_fc(actionFc, linear_grad, rho, fcStr):
    linear_grad = np.array(linear_grad)
    m_hat = sum(linear_grad, 1) / len(actionFc) / rho
    mu_c, mu_d = np.zeros(len(m_hat)), np.zeros(len(m_hat))

    userPrediction = []
    m_norm = np.matmul(np.transpose(m_hat), m_hat)
    for i in range(len(actionFc)):
        if_left = np.matmul(np.transpose(linear_grad[i]), m_hat)
        if if_left > 0.175 * m_norm:
            userPrediction.append(1)
            mu_c += linear_grad[i]
        else:
            userPrediction.append(0)
            mu_d += linear_grad[i]

    userPrediction = np.array(userPrediction)
    # print('User prediction vector: ' + str(userPrediction))
    mu_c = mu_c / sum(userPrediction)
    mu_d = mu_d / (len(actionFc) - sum(userPrediction))
    cov_c = np.zeros((len(m_hat), len(m_hat)))
    cov_d = np.zeros((len(m_hat), len(m_hat)))
    for j in range(len(actionFc)):
        if userPrediction[j] == 1:
            cov_c += np.outer(linear_grad[j] - mu_c, linear_grad[j] - mu_c)
        else:
            cov_d += np.outer(linear_grad[j] - mu_d, linear_grad[j] - mu_d)

    cov_c = cov_c / (sum(userPrediction) - 1)
    cov_d = cov_d / (sum(userPrediction) - 1)

    denom_phi = np.sqrt(np.matmul(np.matmul(np.transpose(mu_c), cov_c), mu_c))
    denom_psi = np.sqrt(np.matmul(np.matmul(np.transpose(mu_d), cov_d), mu_d))
    num_phi = np.matmul(np.transpose(m_hat), mu_c - 0.5 * m_hat)
    num_psi = np.matmul(np.transpose(m_hat), mu_d - 0.5 * m_hat)

    phi = 0.5 + 0.5 * special.erf(num_phi / denom_phi / np.sqrt(2))
    psi = 0.5 - 0.5 * special.erf(-num_psi / denom_psi / np.sqrt(2))

    pi1 = (fcStr[0] * (1 - psi) - fcStr[1] * (1 - phi)) / (phi - psi)
    pi3 = (fcStr[2] * (1 - psi) - fcStr[3] * (1 - phi)) / (phi - psi)
    pi2 = (fcStr[1] * phi - fcStr[0] * psi) / (phi - psi)
    pi4 = (fcStr[3] * phi - fcStr[2] * psi) / (phi - psi)

    for k in range(len(actionFc)):
        if userPrediction[k] == 1:
            if actionFc[k] == 1:
                temp = pi1
            else:
                temp = pi3
        else:
            if actionFc[k] == 1:
                temp = pi2
            else:
                temp = pi4
        rand_num = np.random.rand(1)
        if rand_num < temp:
            actionFc[k] = 1
        else:
            actionFc[k] = 0

    return actionFc, userPrediction


def update_action_fc_quad(actionFc, linear_grad, rho, fcStr):
    linear_grad = np.array(linear_grad)
    m_hat = sum(linear_grad, 1) / len(actionFc) / rho
    # mu_c, mu_d = np.zeros(len(m_hat)), np.zeros(len(m_hat))

    userPrediction = []
    m_norm = np.matmul(np.transpose(m_hat), m_hat)
    for i in range(len(actionFc)):
        if_left = np.matmul(np.transpose(linear_grad[i]), linear_grad[i])
        if if_left > 100:
            userPrediction.append(0)
        else:
            userPrediction.append(1)

    userPrediction = np.array(userPrediction)
    # print('User prediction vector: ' + str(userPrediction))

    # phi = 0.5 + 0.5 * special.erf(num_phi / denom_phi / np.sqrt(2))
    # psi = 0.5 - 0.5 * special.erf(-num_psi / denom_psi / np.sqrt(2))

    pi1 = fcStr[0]
    pi3 = fcStr[2]
    pi2 = fcStr[1]
    pi4 = fcStr[3]

    for k in range(len(actionFc)):
        if userPrediction[k] == 1:
            if actionFc[k] == 1:
                temp = pi1
            else:
                temp = pi3
        else:
            if actionFc[k] == 1:
                temp = pi2
            else:
                temp = pi4
        rand_num = np.random.rand(1)
        if rand_num < temp:
            actionFc[k] = 1
        else:
            actionFc[k] = 0

    return actionFc, userPrediction


def geoMed(grads_, eps=1e-5):
    X = np.array(grads_)
    y = np.mean(X, 0)
    while True:
        D = cdist(X, [y])
        nonzeros = (D != 0)[:, 0]
        # nonzero_indices = np.zeros(nonzeros.shape)
        # for a in range(len(nonzeros)):
        #     if nonzeros[a]:
        #         nonzero_indices[a] = 1

        Dinv = 1 / D[nonzeros]
        Dinvs = np.sum(Dinv)
        W = Dinv / Dinvs
        T = np.sum(W * X[nonzeros], 0)

        num_zeros = len(X) - np.sum(nonzeros)
        if num_zeros == 0:
            y1 = T
        elif num_zeros == len(X):
            return y
        else:
            R = (T - y) * Dinvs
            r = np.linalg.norm(R)
            rinv = 0 if r == 0 else num_zeros / r
            y1 = max(0, 1 - rinv) * T + min(1, rinv) * y

        if euclidean(y, y1) < eps:
            return y1
        y = y1


def update_action_fc_robust(actionFc, linear_grad, m_hat, fcStr):
    linear_grad = np.array(linear_grad)
    # m_hat = geoMed(linear_grad)

    userPrediction = []
    m_norm = np.matmul(np.transpose(m_hat), m_hat)
    for i in range(len(actionFc)):
        if_left = np.matmul(np.transpose(linear_grad[i] - m_hat), linear_grad[i] - m_hat)
        if if_left > 40 * m_norm:
            userPrediction.append(0)
        else:
            userPrediction.append(1)
    userPrediction = np.array(userPrediction)

    for k in range(len(actionFc)):
        if userPrediction[k] == 1:
            if actionFc[k] == 1:
                temp = fcStr[0]
            else:
                temp = fcStr[2]
        else:
            if actionFc[k] == 1:
                temp = fcStr[1]
            else:
                temp = fcStr[3]
        rand_num = np.random.rand(1)
        if rand_num < temp:
            actionFc[k] = 1
        else:
            actionFc[k] = 0

    return actionFc, userPrediction
