import os
import numpy as np
from config import parse_args, save_module
from model import learning_model, update_action_user, update_action_fc, update_action_fc_robust, update_action_fc_quad
from data_loader import data_provider
import torch


def computeProbs(actionUser, userPrediction):
    numC = 0.0000001
    numCC = 0.0000001
    numD = 0.0000001
    numDC = 0.0000001
    for i_ in range(len(actionUser)):
        if actionUser[i_] == 1:
            numC += 1
            if userPrediction[i_] == 1:
                numCC += 1
        else:
            numD += 1
            if userPrediction[i_] == 1:
                numDC += 1
    pDet, pFA = numCC/numC, numDC/numD
    return pDet, pFA


def main(args, save_folder):
    cum_loss = []
    iterator_list = []
    # rho = 1
    for i in range(args.agent_num):
        # Generate predetermined models
        iterator_list.append(data_provider(args))
        print('iterator for ' + str(i))
    # Generate iterator for training accuracy
    iterator_train = data_provider(args, train_='train')
    print('Iterator for training cost is generated.')
    # Generate iterator for testing accuracy
    iterator_test = data_provider(args, train_='test')
    print('Iterator for testing cost is generated.')

    dnn_model = learning_model(args)
    # Perform first iteration

    omega_user = [[args.userStr[0], 1-args.userStr[0], 0, 0], [args.userStr[1], 1-args.userStr[1], 0, 0],
                  [0, 0, args.userStr[2], 1-args.userStr[2]], [0, 0, args.userStr[3], 1-args.userStr[3]]]
    omega_fc = [[args.fcStr[0], 0, 1-args.fcStr[0], 0], [0, args.fcStr[1], 0, 1-args.fcStr[1]],
                [args.fcStr[2], 0, 1-args.fcStr[2], 0], [0, args.fcStr[3], 0, 1-args.fcStr[3]]]
    omega_user, omega_fc = np.array(omega_user), np.array(omega_fc)
    omega = np.matmul(omega_user, omega_fc)

    actionFc = np.ones(args.agent_num)
    actionUser = np.ones(args.agent_num)
    lamb = args.lambda_init

    for t in range(args.iter_num):
        grad_save = []
        loss_save = []
        linear_grad_save = []
        rho = np.matmul(lamb, np.array(args.userStr))
        actionUser = update_action_user(actionUser, actionFc, args.userStr)
        for i_ in range(args.agent_num):
            # Generate data for initial iteration
            data_u, data_v = iterator_list[i_].get_data()
            # Compute gradient
            if args.robust_selection == 'robust':
                grads, cost_, linear_grad = dnn_model.grad_compute_robust(data_u, data_v, t, i_, actionUser[i_])
            else:
                grads, cost_, linear_grad = dnn_model.grad_compute(data_u, data_v, t, i_, actionUser[i_])
            loss_save.append(cost_)
            grad_save.append(grads)
            linear_grad_save.append(linear_grad)
        if args.robust_selection == 'robust':
            m_hat = dnn_model.update_params(grad_save, rho, linear_grad_save)
        else:
            _ = dnn_model.update_params(grad_save, rho)

        if args.robust_selection == 'robust':
            actionFc, userPrediction = update_action_fc_robust(actionFc, linear_grad_save, m_hat, args.fcStr)
        elif args.robust_selection == 'quadratic':
            actionFc, userPrediction = update_action_fc_quad(actionFc, linear_grad_save, rho, args.fcStr)
        else:
            actionFc, userPrediction = update_action_fc(actionFc, linear_grad_save, rho, args.fcStr)
        pDet, pFA = computeProbs(actionUser, userPrediction)
        print('pDet: ' + str(pDet) + ' pFA: ' + str(pFA))
        lamb = np.matmul(lamb, omega)

        # loss_avg = np.sum(loss_save) / args.agent_num
        # print('Iteration: ' + str(t) + ' Average Loss: ' + str(loss_avg))

        if t % args.save_interval == 0:
            batch_train, labels_train = iterator_train.get_data()
            cost_tr, accuracy_tr = dnn_model.check_accuracy(batch_train, labels_train)
            print('Iteration ' + str(t) + ' Training Loss: ' + str(cost_tr) +
                  ' Training Accuracy: ' + str(accuracy_tr))
            save_module(args, save_folder, t, [cost_tr, accuracy_tr],
                        ['training_loss', 'training_accuracy'])

            batch_test, labels_test = iterator_test.get_data()
            cost_test, accuracy_test = dnn_model.check_accuracy(batch_test, labels_test)
            print('Iteration ' + str(t) + ' Testing Loss: ' + str(cost_test) +
                  ' Testing Accuracy: ' + str(accuracy_test))
            save_module(args, save_folder, t, [cost_test, accuracy_test, pFA, pDet],
                        ['testing_loss', 'testing_accuracy', 'pFA', 'pDet'])


if __name__ == '__main__':
    argument = parse_args()
    print(argument.cuda)
    trial = 0
    sim_folder = argument.save_folder + '_' + str(trial)
    while os.path.exists(sim_folder):
        trial += 1
        sim_folder = argument.save_folder + '_' + str(trial)
    os.mkdir(sim_folder)
    setup_file = sim_folder + '/setup.txt'
    f = open(setup_file, "w+")
    f.write(str(argument))
    f.close()
    main(argument, sim_folder)
    print('Finished')
